<?php

use Sunspikes\Ratelimit\Cache\Adapter\DesarrollaCacheAdapter;
use Sunspikes\Ratelimit\Cache\Factory\DesarrollaCacheFactory;
use Sunspikes\Ratelimit\RateLimiter;
use Sunspikes\Ratelimit\Throttle\Factory\ThrottlerFactory;
use Sunspikes\Ratelimit\Throttle\Hydrator\HydratorFactory;

App::uses('AppComponent', 'Controller/Component');

class UserManagerComponent extends AppComponent {

	/**
	 * Settings for rate limiting
	 * @var array
	 */
	public $rateLimit = [
		'attempts' => 5,
		'timeout' => 600
	];

	/**
	 * Set component settings
	 *
	 * @see Component::initialize()
	 */
	public function initialize(Controller $controller) {
		parent::initialize($controller);

		if (!empty($this->settings['rateLimit'])) {
			$this->rateLimit = array_merge($this->rateLimit, $this->settings['rateLimit']);
		}
	}

	/**
	 * register a user
	 *
	 * @param 	object 	$Model 		The model object to use for saving
	 * @return 	array 	$result 	Two element array - result(bool) / data(array)
	 */
	public function register($Model) {
		if (Configure::read('SiteSetting.users.enable_registration')) {
			if ($Model->register($this->_controller->request->data)) {

				$activationType = Configure::read('SiteSetting.users.activation');

				if ($activationType == 'auto') {
					$message = 'Your account has been created, logging in...';

					$fullUser = $Model->getUserForLogin($Model->id);
					$this->_controller->Auth->login($fullUser);
					$redirect = $this->_controller->Auth->redirect();

				} elseif ($activationType == 'email') {
					$message = 'Your account has been created, however you will be unable to login until you have clicked the link in the verifcation email you have been sent.';
				} else {
					$message = 'Your account has been created but needs manually verifying by an administrator. You will be emailed once your account is active.';
				}

				if (! isset($redirect)) {
					$redirect = array(
						'action' => 'login'
					);
				}

				$result = true;
				$returnData = array(
					'redirect' => $redirect,
					'message' => $message
				);

			} else {

				$validationErrors = array();
				foreach ($Model->validationErrors as $key => $errors) {
					$validationErrors[] = $errors['0'];
				}

				$errors = array(
					'description' => 'Please correct the following issues and try again',
					'list' => $validationErrors
				);

				$result = false;
				$returnData = array(
					'errors' => $errors
				);
			}
		} else {
			$result = false;
			$returnData = array(
				'errors' => 'Registration is disabled'
			);
		}

		return array(
			'result' => $result,
			'data' => $returnData
		);
	}

	/**
	 * attempt to login a user
	 *
	 * @param 	object 	$Model 		The model object to use for saving
	 * @return 	array 	$result 	Two element array - result(bool) / data(array)
	 */
	public function login($Model) {
		// Setup the throttler, this stops someone continuously trying usernames from the same IP
		$loginThrottler = $this->_getLoginThrottler();

		$Model->validate = $Model->validateLogin();

		$Model->set($this->_controller->request->data);

		// Check if the IP has been temporarily banned
		if (!$loginThrottler->check()) {
			// there is a delay preventing login

			$result = false;
			$returnData = array(
				'errors' => array(
					'title' => __('Login failed'),
					'description' => __('You have failed to login %s times, You will not be able to login for the next %s minutes!', $loginThrottler->getLimit(), $loginThrottler->getTtl() / 60)
				)
			);
		} else {
			// no login delay set
			// proceed as normal

			if ($Model->validates()) {
				if ($this->_controller->Auth->login()) {
					// Unset the login attempt limits

					$this->_controller->Session->write('Login.attempt', 0);
					$this->_controller->Session->write('Login.delay', 0);

					//Get a better version of the user model out in the format $data['User']
					$user = $this->_controller->Auth->user();

					$fullUser = $Model->getUserForLogin($user['id']);

					$this->_controller->Auth->login($fullUser);

					$result = true;
					$returnData = array(
						'redirect' => $this->_controller->Auth->redirect(),
						'message' => 'Login successful'
					);
				} else {
					// Define the time delays for limiting login attempts
					$loginDelays = array(0, 1, 5, 30, 60, 90);

					// Check login limits and write updated ones to session
					$loginThrottler->hit();
					// $loginAttempt = $this->_controller->Session->read('Login.attempt') ? $this->_controller->Session->read('Login.attempt') + 1 : 1;
					// $loginDelay = isset($loginDelays[intval($loginAttempt / 5)]) ? $loginDelays[intval($loginAttempt / 5)] : 90;
					// $this->_controller->Session->write('Login.attempt', $loginAttempt);
					// $this->_controller->Session->write('Login.delay', time() + $loginDelay * 60);

					$result = false;

					$returnData = array(
						'errors' => array(
							'title' => __('Login failed'),
							'description' => __('Username or password is incorrect')
						)
					);
				}

			} else {
				// validation failed
				$validationErrors = array();
				foreach ($Model->validationErrors as $key => $errors) {
					$validationErrors[] = $errors['0'];
				}

				$result = false;
				$returnData = array(
					'errors' => array(
						'description' => 'The email and / or password was incorrect',
						'list' => $validationErrors
					)
				);

				unset($this->_controller->request->data[$Model->alias]['password']);
			}
		}

		return array(
			'result' => $result,
			'data' => $returnData
		);
	}

	/**
	 * Sets up the login throttler to stop users brute forcing from the same IP address.
	 * @return mixed A rate throttler for the login using the client IP as the unique key
	 */
	protected function _getLoginThrottler() {
		$config = [
			'driver' => 'file',
			'file' => [
				'cache_dir' => APP . DS . 'tmp' . DS . 'rate_limiting'
			]
		];
		$cacheAdapter = new DesarrollaCacheAdapter((new DesarrollaCacheFactory(null, $config))->make());

		$ratelimiter = new RateLimiter(new ThrottlerFactory(), new HydratorFactory(), $cacheAdapter, $this->rateLimit['attempts'], $this->rateLimit['timeout']);

		// 2. Get a throttler for key login and the client IP
		$loginThrottler = $ratelimiter->get(['login', $this->_controller->request->clientIp()]);

		return $loginThrottler;
	}

}
